# -*- coding: utf-8 -*-
import pkg_resources
from datetime import timedelta
import logging

from bos_mint import Config

from .node import Node, NodeException
from . import tostring, __VERSION__
from . import datestring

from .dataproxy_link.ping import Ping

from flask import (
    flash
)

CACHE_VERSIONS = None
CACHE_ACCOUNTS = None


def clear_accounts_cache():
    global CACHE_ACCOUNTS
    CACHE_ACCOUNTS = None


def getMenuInfo():
    info_less = Config.get("notifications", "accountLessThanCoreInfo", 1000)
    warning_less = Config.get("notifications", "accountLessThanCoreWarning", 200)

    try:
        witnessAccount = Node().getWitnessAccount()
        for balance in witnessAccount.balances:
            if balance.asset["id"] == "1.3.0":
                if float(balance) < warning_less:
                    flash("Account witness-account has only " + str(float(balance)) + " " + balance.asset["symbol"] + ", please replenish immediately", category="warning")
                elif float(balance) < info_less:
                    flash("Account witness-account has only " + str(float(balance)) + " " + balance.asset["symbol"] + ", please replenish")
    except NodeException:
        pass

    try:
        account = Node().getSelectedAccount()
        for balance in account.balances:
            if balance.asset["id"] == "1.3.0":
                if float(balance) < warning_less:
                    flash("Account " + account.name + " has only " + str(float(balance)) + " " + balance.asset["symbol"] + ", please replenish immediately", category="warning")
                elif float(balance) < info_less:
                    flash("Account " + account.name + " has only " + str(float(balance)) + " " + balance.asset["symbol"] + ", please replenish")
        accountDict = {
            'id': account.identifier,
            'name': account.name,
            'toString': tostring.toString(account)}
    except NodeException:
        try:
            any_account = len(Node().getAllAccountsOfWallet()) > 0
        except Exception:
            any_account = False
        if any_account:
            accountDict = {'id': '-', 'name': '-', 'toString': 'Please select an account'}
        else:
            accountDict = {'id': '-', 'name': '-', 'toString': 'Please add an account'}

    numberOfOpenTransactions = 0
    try:
        currentTransaction = Node().getPendingTransaction()
        if currentTransaction:
            numberOfOpenTransactions = len(Node().getPendingTransaction().ops)
    except NodeException:
        pass

    numberOfVotableProposals = 0
    try:
        numberOfVotableProposals = len(Node().getAllProposals())
    except NodeException:
        pass

    walletLocked = True
    try:
        walletLocked = Node().locked()
    except Exception:
        pass

    global CACHE_VERSIONS
    if CACHE_VERSIONS is None:
        CACHE_VERSIONS = {}
        for name in ["bos-incidents", "peerplays", "bookiesports"]:
            try:
                CACHE_VERSIONS[name] = pkg_resources.require(name)[0].version
            except pkg_resources.DistributionNotFound:
                CACHE_VERSIONS[name] = "not installed"

    menuInfo = {
        'account': accountDict,
        'numberOfOpenTransactions': numberOfOpenTransactions,
        'numberOfVotableProposals': numberOfVotableProposals,
        'walletLocked': walletLocked,
        'version': __VERSION__,
        'versions': CACHE_VERSIONS
    }

    global CACHE_ACCOUNTS
    if CACHE_ACCOUNTS is None:
        CACHE_ACCOUNTS = []
        try:
            for account in Node().getAllAccountsOfWallet():
                if account['name']:
                    CACHE_ACCOUNTS.append({
                        'id': account['account'].identifier,
                        'name': account['account'].name,
                        'publicKey': account['pubkey'],
                        'toString': account['account'].identifier + ' - ' + account['account'].name})
                else:
                    CACHE_ACCOUNTS.append({
                        'id': 'None',
                        'name': 'This shouldnt happen',
                        'publicKey': 'None',
                        'toString': 'None - Error shouldnt happen' + account['pubkey']})
        except NodeException:
            pass

    menuInfo['allAccounts'] = CACHE_ACCOUNTS

    try:
        menuInfo['chain'] = {
            "name": Config.get("connection", "use"),
            "id": Node().get_node().rpc.chain_params["chain_id"],
            "block": Node().get_node().rpc.get_object("2.1.0")["head_block_number"],
            "time": Node().get_node().rpc.get_object("2.1.0")["time"] + "Z"
        }
    except Exception as e:
        menuInfo['chain'] = {
            "name": Config.get("connection", "use"),
            "id": str(e),
            "block": "-",
            "time": datestring.date_to_string()
        }

    if (datestring.string_to_date() - timedelta(seconds=30) > datestring.string_to_date(menuInfo["chain"]["time"])):
        menuInfo["chain"]["out_of_sync"] = True

    try:
        menuInfo['incidents'] = {
            "dataproxy_link": Ping().get_status()
        }
    except Exception:
        pass

    menuInfo["advanced_features"] = Config.get("advanced_features", False)

    logging.getLogger(__name__).debug("getMenuInfo done")

    return menuInfo



# Global to string method for centralized visualization
from .istring import InternationalizedString


def findEnglishOrFirst(listOfIStrings, desiredLanguage='en'):
    return InternationalizedString.listToDict(listOfIStrings).get(
        desiredLanguage, listOfIStrings[0][1])


def toString(toBeFormatted, object=None):

    if type(object).__name__ == "Event":
        if isinstance(toBeFormatted.get('name'), list):
            name = findEnglishOrFirst(toBeFormatted.get('name'))
        else:
            name = toBeFormatted.get('name')
        if name is None:
            displayName = "Unknown (deleted?)"
        else:
            displayName = name
        if toBeFormatted.get("start_time", None) is not None:
            displayName = str(toBeFormatted["start_time"]) + "Z" + ":\n" + displayName
        if toBeFormatted.get("id", None) is not None:
            displayName = displayName + ' (' + toBeFormatted['id'] + ')'
    elif toBeFormatted.get('name') and toBeFormatted.get('id'):
        if isinstance(toBeFormatted.get('name'), list):
            name = findEnglishOrFirst(toBeFormatted.get('name'))
        else:
            name = toBeFormatted.get('name')

        displayName = name + ' (' + toBeFormatted.get('id') + ')'
    elif toBeFormatted.get('description') and toBeFormatted.get('id'):
        if isinstance(toBeFormatted.get('description'), list):
            name = findEnglishOrFirst(toBeFormatted.get('description'))
        else:
            name = toBeFormatted.get('description')

        displayName = name + ' (' + toBeFormatted.get('id') + ')'
    elif toBeFormatted.get('id'):
        displayName = '(' + toBeFormatted.get('id') + ')'
        if object:
            displayName = object.__class__.__name__ + " " + displayName
    else:
        raise Exception

    if toBeFormatted.get('pendingOperationId'):
        displayName = displayName + '*(' +\
            toBeFormatted.get('pendingOperationId') + ')'

    return displayName



# -*- coding: utf-8 -*-
from flask import redirect, flash, url_for, request, render_template
from functools import wraps
from datetime import datetime
from peerplaysbase.operationids import getOperationNameForId

from .node import Node, NodeException
from . import wrapper
from . import config
from . import datestring
from .menu_info import getMenuInfo

# dictionary to configure types (Sport, EventGroup, etc.)
#  title: human readable title
TYPENAMES = {
    'sport': {'title': 'Sport'},
    'eventgroup': {'title': 'Event group'},
    'event': {'title': 'Event'},
    'event_status': {'title': 'Event (only status)'},
    'bettingmarketgroup': {'title': 'Betting market group'},
    'bettingmarket': {'title': 'Betting market'},
    'bet': {'title': 'Bet'},
    'bettingmarketgrouprule': {'title': 'Betting market group rule'}
}

# get list of objects for typename, containing id, typeName and toString field
TYPE_GET_ALL = {
    'sport': lambda unusedId: [
        wrapper.Sport(**x) for x in Node().getSports()
    ],
    'eventgroup': lambda tmpSportId: [
        wrapper.EventGroup(**x) for x in Node().getEventGroups(tmpSportId)
    ],
    'event': lambda tmpEventGroupId: [
        wrapper.Event(**x) for x in Node().getEvents(tmpEventGroupId)
    ],
    'event_status': lambda tmpEventGroupId: [
        wrapper.Event(**x) for x in Node().getEvents(tmpEventGroupId)
    ],
    'bettingmarketgroup': lambda tmpEventId: [
        wrapper.BettingMarketGroup(**x) for x in Node().getBettingMarketGroups(tmpEventId)
    ],
    'bettingmarket': lambda tmpBMGId: [
        wrapper.BettingMarket(**x) for x in Node().getBettingMarkets(tmpBMGId)
    ],
    'bettingmarketgrouprule': lambda unusedId: [
        wrapper.BettingMarketGroupRule(**x) for x in Node().getBettingMarketGroupRules() if x is not None
    ],
    'bet': lambda tmpBMGId: [],  # not implemented yet
}

# get list of objects for typename, containing id, typeName and toString field
TYPE_GET = {
    'sport': lambda tmpId: Node().getSport(tmpId),
    'eventgroup': lambda tmpId: Node().getEventGroup(tmpId),
    'event': lambda tmpId: Node().getEvent(tmpId),
    'event_status': lambda tmpId: Node().getEvent(tmpId),
    'bettingmarketgroup': lambda tmpId: Node().getBettingMarketGroup(tmpId),
    'bettingmarketgrouprule': lambda tmpId: Node().getBettingMarketGroupRule(tmpId),
    'bettingmarket': lambda tmpId: Node().getBettingMarket(tmpId),
    'bet': lambda tmpId: None,  # not implemented yet
}

# get object for typename, containing id of parent object
PARENTTYPE_GET = {
    'sport': lambda tmpId: None,
    'eventgroup': lambda tmpId: Node().getEventGroup(tmpId).sport.identifier,
    'event': lambda tmpId: Node().getEvent(tmpId).eventgroup.identifier,
    'event_status': lambda tmpId: Node().getEvent(tmpId).eventgroup.identifier,
    'bettingmarketgroup': lambda tmpId: Node().getBettingMarketGroup(tmpId).event.identifier,
    'bettingmarketgrouprule': lambda tmpId: None,
    'bettingmarket': lambda tmpId: Node().getBettingMarket(tmpId).bettingmarketgroup.identifier,
    'bet': lambda tmpId: tmpId,
}

# which type does it cascade to
CHILD_TYPE = {
    'sport': 'eventgroup',
    'eventgroup': 'event',
    'event': 'bettingmarketgroup',
    'event_status': 'bettingmarketgroup',
    'bettingmarketgroup': 'bettingmarket',
    'bettingmarketgrouprule': None,
    'bettingmarket': 'bet'
}

# which type does it come from
PARENT_TYPE = {
    'sport': None,
    'eventgroup': 'sport',
    'event': 'eventgroup',
    'bettingmarketgroup': 'event',
    'bettingmarketgrouprule': None,
    'bettingmarket': 'bettingmarketgroup',
    'bet': 'bettingmarket'
}

TYPENAME_TO_NEWOP_MAP = {
    'sport': 'sport_create',
    'eventgroup': 'event_group_create',
    'event': 'event_create',
    'event_status': None,
    'bettingmarketgroup': 'betting_market_group_create',
    'bettingmarketgrouprule': 'betting_market_rules_create',
    'bettingmarket': 'betting_market_create',
    'bet': 'bet_create',
}

TYPENAME_TO_UPDATEOP_MAP = {
    'sport': 'sport_update',
    'eventgroup': 'event_group_update',
    'event': 'event_update',
    'event_status': 'event_update_status',
    'bettingmarketgroup': 'betting_market_group_update',
    'bettingmarketgrouprule': 'betting_market_rules_update',
    'bettingmarket': 'betting_market_update'
}

UPDATEOP_TO_TYPENAME_MAP = {
    'sport_update': 'sport',
    'event_group_update': 'eventgroup',
    'event_update': 'event',
    'event_update_status': 'event_status',
    'betting_market_group_update': 'bettingmarketgroup',
    'betting_market_rules_update': 'bettingmarketgrouprule',
    'betting_market_update': 'bettingmarket',
    'betting_market_group_resolve': 'bettingmarketgroup'
}


def filterOnlyAllowed(enumClazz, currentStatus):
    allowed = config.get("allowed_transitions", {}).get(enumClazz.__name__, {})

    return [(x, x) for x in enumClazz.options if "COUNT" not in x and (x in allowed.get(currentStatus, []) or x == currentStatus)]


def toString(toBeFormatted):
    raise Exception


def getComprisedParentByIdGetter(typeName):
    def doGet(parentId):
        # unifies objects on the blockchain and in the cache
        if parentId and parentId.startswith('0.0'):
            return getParentByIdFromPendingProposalGetter(typeName)(parentId)
        else:
            return getParentByIdGetter(typeName)(parentId)

    return doGet


def getParentByIdFromPendingProposalGetter(typeName):
    parentTypeName = PARENT_TYPE.get(typeName)

    def doGet(parentId):
        bufferedObjects = getProposalOperations(Node().get_node().tx())

        for bufferedObject in bufferedObjects:
            if bufferedObject['typeName'] == parentTypeName\
                    and bufferedObject['id'] == parentId:
                return bufferedObject

        raise NodeException("There was no creation operation found for a " +
                            PARENT_TYPE.get(typeName) + " with identifier=" +
                            str(parentId))

    return doGet


def getParentByIdGetter(typeName):
    parentTypeName = PARENT_TYPE.get(typeName)
    if parentTypeName:
        return TYPE_GET.get(parentTypeName)

    return None


def getChildType(typeName):
    return CHILD_TYPE.get(typeName)


def getParentType(typeName):
    return PARENT_TYPE.get(typeName)


def getParentTypeGetter(typeName):
    return PARENTTYPE_GET.get(typeName)


def getComprisedParentTypeGetter(typeName):
    def doGet(selectedId):
        # unifies objects on the blockchain and in the cache
        if selectedId and selectedId.startswith('0.0'):
            bufferedObjects = getProposalOperations(Node().get_node().tx())
            for bufferedObject in bufferedObjects:
                if bufferedObject['id'] == selectedId:
                    return bufferedObject['parentId']
            else:
                return selectedId
        else:
            return getParentTypeGetter(typeName)(selectedId)

    return doGet


def getTypeGetter(typeName):
    return TYPE_GET.get(typeName)


def getTypesGetter(typeName):
    return TYPE_GET_ALL.get(typeName)


def getComprisedTypesGetter(typeName):

    def doGet(parentId):
        # unifies objects on the blockchain and in the cache
        if parentId and parentId.startswith('0.0'):
            get1 = []
        else:
            get1 = getTypesGetter(typeName)(parentId)

        get2 = getTypesFromPendingProposalGetter(typeName)(parentId)

        # allow cache to override
        bufferDict = {item['id']: item for item in get1 + get2}
        return bufferDict.values()

    return doGet


def getTypesFromPendingProposalGetter(typeName):
    def doGet(parentId):
        inBuffer = []

        bufferedObjects = getProposalOperations(Node().get_node().tx())

        for bufferedObject in bufferedObjects:
            if bufferedObject['typeName'] == typeName and\
                    (not parentId or bufferedObject['parentId'] == parentId or not bufferedObject['parentId']):
                inBuffer.append(bufferedObject)

        return inBuffer

    return doGet


def getTitle(typeName):
    # insert configurable translation here
    return TYPENAMES[typeName]['title']


def strfdelta(time, fmt):
    if not hasattr(time, "days"):  # dirty hack
        now = datetime.now()
        if isinstance(time, str):
            time = datetime.strptime(time, '%Y-%m-%dT%H:%M:%S')
        time = abs(now - time)

    d = {"days": time.days}
    d["hours"], rem = divmod(time.seconds, 3600)
    d["minutes"], d["seconds"] = divmod(rem, 60)
    return fmt.format(**d)


def wallet_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        if not Node().wallet_exists():
            flash(
                "No wallet has been created yet. You need to do that and "
                "import your witness active key"
            )
            return redirect(url_for('newwallet', next=request.url))
        return f(*args, **kwargs)
    return decorated_function


def unlocked_wallet_required(f):
    @wraps(f)
    @wallet_required
    def decorated_function(*args, **kwargs):
        if Node().locked():
            flash(
                "In order to access this functionality, you need "
                "to unlock your wallet first!"
            )
            return redirect(url_for('unlock', next=request.url))
        return f(*args, **kwargs)
    return decorated_function


def render_template_menuinfo(tmpl_name, **kwargs):
    """
    If there is a general error in rendering, simply return a 500 error.
    Must be robust to connection lost.

    :param tmpl_name: name of the template to be rendered
    :type tmpl_name: str
    """
    menuInfo = getMenuInfo()
    return render_template(tmpl_name, menuInfo=menuInfo, **kwargs)


def processNextArgument(nextArg, default):
    if not nextArg:
        return url_for(default)

    if nextArg.startswith('/'):
        if nextArg:
            nextWords = nextArg.split(sep='/')

            if nextWords[1] == 'overview':
                try:
                    return url_for('overview',
                                   typeName=nextWords[2],
                                   identifier=nextWords[3])
                except Exception:
                    return url_for(default)
    else:
        return nextArg


def isProposal(res):
    if type(res).__name__ == "ProposalBuilder":
        return True
    else:
        return res['operations'][0][0] == 22


def getProposalOperations(tx):
    convertedOperations = []

    if tx.get('operations', None):
        if isProposal(tx):
            # its a proposal, proceed
            for idx, operation in enumerate(
                tx['operations'][0][1]['proposed_ops']
            ):
                operation = operation['op']
                operationName = getOperationNameForId(operation[0])
                typeName = 'unknown'
                for tmpTypeName, newOpName in TYPENAME_TO_NEWOP_MAP.items():
                    if newOpName == operationName:
                        typeName = tmpTypeName

                typeName = UPDATEOP_TO_TYPENAME_MAP.get(operationName, typeName)

                if typeName == 'unknown':
                    raise Exception
                # this is a hack. proper id construction __must__ be
                # provided by backend
                operationId = '0.0.' + str(idx)
                tmpData = operation[1].copy()
                tmpData.update({'operationName': operationName,
                                'operationId': operationId})

                bufferedObject = wrapper.mapOperationToObject(
                    typeName, tmpData)
                convertedOperations.append(bufferedObject)

    return convertedOperations


def date_to_string(date_object=None):
    return datestring.date_to_string(date_object)


def string_to_date(date_string=None):
    return datestring.string_to_date(date_string)



import os
import yaml
import pprint
import logging
import pkg_resources

from flask import Flask, jsonify, redirect, url_for, flash
from flask_sqlalchemy import SQLAlchemy
from logging.handlers import TimedRotatingFileHandler
from werkzeug.exceptions import HTTPException, InternalServerError
from peerplays.instance import set_shared_config
import collections
from copy import deepcopy
import io


def get_version():
    try:
        with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'VERSION')) as version_file:
            return version_file.read().strip()
    except FileNotFoundError:
        with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", 'VERSION')) as version_file:
            return version_file.read().strip()


__VERSION__ = get_version()


class Config():
    """ This class allows us to load the configuration from a YAML encoded
        configuration file.
    """

    ERRORS = {
        "secret_key": "Please create a configuration file config-bos-mint.yaml in your working directory with a secret_key entry, see config-example.yaml",
        "connection.use": "Please create a configuration file config-bos-mint.yaml in your working directory with a connection.use entry, see config-example.yaml"
    }

    data = None
    source = None

    @staticmethod
    def load(config_files=[], relative_location=False):
        """ Load config from a file

            :param str file_name: (defaults to ['config.yaml']) File name and
                path to load config from
        """

        if not Config.data:
            Config.data = {}

        if not config_files:
            # load all config files as default
            config_files.append("config-defaults.yaml")
        if type(config_files) == str:
            config_files = [config_files]

            for config_file in config_files:
                if relative_location:
                    file_path = config_file
                else:
                    file_path = os.path.join(
                        os.path.dirname(os.path.realpath(__file__)),
                        config_file
                    )
                stream = io.open(file_path, 'r', encoding='utf-8')
                with stream:
                    Config.data = Config._nested_update(Config.data, yaml.safe_load(stream))

            if not Config.source:
                Config.source = ""
            Config.source = Config.source + ";" + ";".join(config_files)

    @staticmethod
    def get_config(config_name=None):
        """ Static method that returns the configuration as dictionary.
            Usage:

            .. code-block:: python

                Config.get_config()
        """
        if not config_name:
            if not Config.data:
                raise Exception("Either preload the configuration or specify config_name!")
        else:
            if not Config.data:
                Config.data = {}
            Config.load(config_name)
        return deepcopy(Config.data)

    @staticmethod
    def get(*args, **kwargs):
        """
        This config getter method allows sophisticated and encapsulated access to the config file, while
        being able to define defaults in-code where necessary.

        :param args: key to retrieve from config, nested in order. if the last is not a string it is assumed to be the default, but giving default keyword is then forbidden
        :type tuple of strings, last can be object
        :param message: message to be displayed when not found, defaults to entry in ERRORS dict with the
                                key defined by the desired config keys in args (key1.key2.key2). For example
                                Config.get("foo", "bar") will attempt to retrieve config["foo"]["bar"], and if
                                not found raise an exception with ERRORS["foo.bar"] message
        :type message: string
        :param default: default value if not found in config
        :type default: object
        """
        default_given = "default" in kwargs
        default = kwargs.pop("default", None)
        message = kwargs.pop("message", None)
        # check if last in args is default value
        if type(args[len(args) - 1]) != str:
            if default_given:
                raise KeyError("There can only be one default set. Either use default=value or add non-string values as last positioned argument!")
            default = args[len(args) - 1]
            args = args[0:len(args) - 1]

        try:
            nested = Config.data
            for key in args:
                if type(key) == str:
                    nested = nested[key]
                else:
                    raise KeyError("The given key " + str(key) + " is not valid.")
            if nested is None:
                raise KeyError()
        except KeyError:
            lookup_key = '.'.join(str(i) for i in args)
            if not message:
                if Config.ERRORS.get(lookup_key):
                    message = Config.ERRORS[lookup_key]
                else:
                    message = "Configuration key {0} not found in {1}!"
                message = message.format(lookup_key, Config.source)
            if default_given is not None:
                logging.getLogger(__name__).debug(message + " Using given default value.")
                return default
            else:
                raise KeyError(message)

        return nested

    @staticmethod
    def reset():
        """ Static method to reset the configuration storage
        """
        Config.data = None
        Config.source = None

    @staticmethod
    def _nested_update(d, u):
        for k, v in u.items():
            if isinstance(v, collections.Mapping):
                d[k] = Config._nested_update(d.get(k, {}), v)
            else:
                if d:
                    d[k] = v
                else:
                    d = {}
                    d[k] = v
        return d


if not Config.data:
    Config.load("config-defaults.yaml")
    notify = False
    try:
        # overwrites defaults
        Config.load("config-bos-mint.yaml", True)
        notify = True
    except FileNotFoundError:
        pass

    if notify:
        logging.getLogger(__name__).info("Custom config has been loaded " + Config.source)


def get_config():
    Config.load("config-defaults.yaml")
    notify = False
    try:
        # overwrites defaults
        Config.load("config-bos-mint.yaml", True)
        notify = True
    except FileNotFoundError:
        pass

    if notify:
        logging.getLogger(__name__).info("bos-mint: Custom config has been loaded " + Config.source)

    Config.get("connection", "use")
    Config.get("secret_key")

    config = Config.data

    config["sql_database"] = config["sql_database"].format(cwd=os.getcwd())

    return config


def set_global_logger():
    # setup logging
    log_folder = os.path.join("dump", "logs")
    log_format = ('%(asctime)s %(levelname) -10s: %(message)s')
    os.makedirs(log_folder, exist_ok=True)
    trfh = TimedRotatingFileHandler(
        os.path.join(log_folder, "manual_intervention.log"),
        "midnight",
        1
    )
    trfh.suffix = "%Y-%m-%d"
    trfh.setFormatter(logging.Formatter(log_format))

    sh = logging.StreamHandler()
    sh.setFormatter(logging.Formatter(log_format))

    # set global logger (e.g. for werkzeug)
    if Config.get("debug"):
        trfh.setLevel(logging.DEBUG)
        sh.setLevel(logging.DEBUG)
        logging.basicConfig(level=logging.DEBUG,
                            format=log_format,
                            handlers=[trfh, sh])
    else:
        trfh.setLevel(logging.INFO)
        sh.setLevel(logging.INFO)
        logging.basicConfig(level=logging.INFO,
                            format=log_format,
                            handlers=[trfh, sh])

    return [trfh, sh]


def set_flask_logger(flask_app, handlers):
    # set app.logger from flask
    while len(app.logger.handlers) > 0:
        app.logger.removeHandler(app.logger.handlers[0])

    for handler in log_handlers:
        app.logger.addHandler(handler)


def set_app_config(flask_app, config):
    basedir = os.path.abspath(os.path.dirname(__file__))

    # Flask Settings
    flask_app.config['DEBUG'] = config["debug"]
    flask_app.config['SECRET_KEY'] = config["secret_key"]

    # Let's store the whole config struct
    flask_app.config['PROJECT'] = config

    # Config database
    flask_app.config['SQLALCHEMY_DATABASE_URI'] = config["sql_database"].format(cwd=basedir)
    flask_app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    flask_app.config['SQLALCHEMY_ECHO'] = config["debug"]

    # disable CSRF protection for now, fix and remove
    flask_app.config['WTF_CSRF_ENABLED'] = False


def set_peerplays_connection(config):
    use = config["connection"]["use"]
    connection_config = config["connection"][use]

    # avoid connectivity loops
    if connection_config.get("num_retries", None) is None:
        connection_config["num_retries"] = 3

    set_shared_config(connection_config)


def set_error_handling(flask_app):
    def handle_exception(e):
        if isinstance(e, HTTPException):
            raise e

        flask_app.logger.exception(e)
        flash(e.__class__.__name__ + ": " + str(e), category='error')
        return redirect(url_for('overview'))

    flask_app.errorhandler(Exception)(handle_exception)
    # In the case of an internal error after the user function
    # (e.g. a view does not return any value)
    # flask is looking for a specifically set InternalServerError handler,
    # see flask/app.py:handle_exception:1532
    flask_app.errorhandler(InternalServerError)(handle_exception)


config = get_config()

log_handlers = set_global_logger()

# Instanciate flask
app = Flask(__name__)

set_flask_logger(app, log_handlers)
set_app_config(app, config)
set_error_handling(app)
db = SQLAlchemy(app)
set_peerplays_connection(config)


app.logger.debug(pprint.pformat(config))


@app.before_first_request
def before_first_request():
    try:
        db.create_all()
    except Exception as e:
        app.logger.warning(str(e))


@app.teardown_appcontext
def shutdown_session(exception=None):
    db.session.remove()


from flask import (
    redirect,
    request,
    session,
    flash,
    url_for,
    abort,
    jsonify,
    send_from_directory
)
from wtforms import FormField, SubmitField
from peerplays.exceptions import WalletExists
import requests
import re

from . import app, forms, utils, widgets, Config
from .menu_info import clear_accounts_cache
from .forms import (
    TranslatedFieldForm,
    NewWalletForm,
    GetAccountForm,
    ReplayForm,
    ApprovalForm,
    BettingMarketGroupResolveForm,
    SynchronizationForm
)
from .models import (
    LocalProposal,
    ViewConfiguration
)
from .node import (
    Node,
    NodeException,
    BroadcastActiveOperationsExceptions
)
from .utils import (
    render_template_menuinfo,
    unlocked_wallet_required,
    wallet_required
)
from .dataproxy_link.ping import Ping

import os
from bos_incidents import factory
from bos_incidents.exceptions import EventNotFoundException,\
    IncidentStorageLostException

from bookiesports import BookieSports
from strict_rfc3339 import InvalidRFC3339Error
from bos_mint.istring import InternationalizedString
from datetime import timedelta
from bos_incidents.format import get_reconstruct_string
from time import sleep
from threading import Thread
from datetime import datetime
from peerplays.event import Events
from bos_mint.datestring import string_to_date
from pprint import pformat, pprint
import json


###############################################################################
# Homepage
###############################################################################


@app.route('/favicon.ico')
def favicon():
    return send_from_directory(os.path.join(app.root_path, 'static'),
                               'favicon.ico', mimetype='image/vnd.microsoft.icon')


@app.route('/')
def index():
    return redirect(url_for('overview'))


@app.route('/lock', methods=['GET', 'POST'])
@wallet_required
def lock():
    Node().lock()
    return redirect(url_for('overview'))


@app.route('/unlock', methods=['GET', 'POST'])
@wallet_required
def unlock():
    unlockForm = forms.UnlockForm()

    if unlockForm.validate_on_submit():
        return redirect(utils.processNextArgument(
            request.args.get('next'), 'overview'))

    return render_template_menuinfo('unlock.html', **locals())


@app.route('/account/info', methods=['GET', 'POST'])
@unlocked_wallet_required
def account_info():
    form = forms.AccountForm()

    if form.validate_on_submit():
        if Config.get("advanced_features", default=False):
            ViewConfiguration.set('synchronous_operations', 'enabled', form.synchronous_operations.data)
            Node().get_node().blocking = ViewConfiguration.get('synchronous_operations', 'enabled', False)
            message = "Synchronous operations is set to " + str(ViewConfiguration.get('synchronous_operations', 'enabled', False))
            if ViewConfiguration.get('synchronous_operations', 'enabled', False):
                message = message + " Calls are now blocking."
            flash(message)
        else:
            flash("Updating account settings is currently disabled")
    else:
        account = Node().getSelectedAccount()
        form.fill(account)
        form.synchronous_operations.data = ViewConfiguration.get('synchronous_operations', 'enabled', False)

    return render_template_menuinfo("account.html", **locals())


@app.route('/bookiesports/sync', methods=['GET', 'POST'])
@unlocked_wallet_required
def bookiesports_sync():
    if not Config.get("advanced_features", default=False):
        raise Exception("Sync currently disabled")

    form = SynchronizationForm()

    if form.validate_on_submit():
        flash("Blockchain has been synchronized with bookiesports and a new proposal has been created, ready to be broadcasted")
        Node().sync(Config.get("connection", "use"))
        return redirect(url_for("pending_operations"))
    else:
        form.chain_id.data = Config.get("connection", "use") + ":" + Node().get_node().rpc.chain_params["chain_id"]
        form.bookiesports_name.data = Config.get("connection", "use")
        try:
            bookieSports = BookieSports(
                chain=form.bookiesports_name.data
            )
            form.bookiesports_name.data = form.bookiesports_name.data + ":" + bookieSports.index.get("chain_id", None)

            if bookieSports.index.get("chain_id", None) is not None and\
                (bookieSports.index["chain_id"] == "*" or
                 bookieSports.index["chain_id"] == Node().get_node().rpc.chain_params["chain_id"]):
                if Node().isInSync(Config.get("connection", "use")):
                    form.status.data = "In sync!"
                else:
                    form.status.data = "NOT in sync! Needs syncing!"
            else:
                flash("Your bookiesports name and chain id do not match. Please make sure that the alias you use for the key connection.use within config-bos-mint.yaml matches the corresponding chain", category="error")
                form.status.data = "Your bookiesports name and chain id do not match."
                delattr(form, "submit")

        except AssertionError:
            flash("BookieSports uses the aliases " + str(BookieSports.list_chains()) + ", please use the same aliases in the key connection.use within config-bos-mint.yaml", category="error")
            delattr(form, "submit")

    return render_template_menuinfo('generic.html', **locals())


@app.route("/account/select/<accountId>", methods=['GET', 'POST'])
@wallet_required
def account_select(accountId):
    try:
        accountName = Node().selectAccount(accountId)
        flash('Account ' + accountName + ' selected!')
    except BroadcastActiveOperationsExceptions as e:
        flash(str(e), category='error')
        return redirect(url_for('pending_operations'))

    return redirect(utils.processNextArgument(
        request.args.get('next'), 'overview'))


@app.route("/account/add", methods=['GET', 'POST'])
@unlocked_wallet_required
def account_add():
    form = GetAccountForm()
    formTitle = "Add Account to wallet"

    if form.validate_on_submit():
        # submit checks if account exists
        try:
            Node().addAccountToWallet(form.privateKey.data)
            clear_accounts_cache()
            flash("Key and all underlying registered accounts imported!")
        except Exception as e:
            flash(
                'There was a problem adding the account to the wallet. ({})'.format(
                    str(e)),
                category='error')

        return redirect(url_for('overview'))

    return render_template_menuinfo('generic.html', **locals())


@app.route("/incidents/replay", methods=['GET', 'POST'])
@app.route("/incidents/replay/<filter>", methods=['GET', 'POST'])
@app.route("/incidents/replay/<filter>/<use>", methods=['GET', 'POST'])
@unlocked_wallet_required
def replay(filter=None, use="mongodb"):
    if use == "bos-auto":
        use = "mongodb"
    if not Config.get("advanced_features", False):
        abort(404)

    form = ReplayForm()
    form.chain.data = Config.get("connection", "use")
    formTitle = "Replay incidents"
    formMessages = []

    proxy_incidents = None

    if not form.back.data and form.validate_on_submit():
        if form.check.data:
            from .dataproxy_link import control
            # query dataproxy
            try:
                proxy_incidents = control.get_replayable_incidents(form.unique_string.data, form.dataproxy.data, form.chain.data, form.witness.data)

                form.replay.render_kw = {'disabled': False}

                del form.check

                formMessages.append(str(len(proxy_incidents)) + " incidents found on dataproxy")
                for _tmp in proxy_incidents:
                    if type(_tmp) == dict:
                        formMessages.append(" - " + _tmp["unique_string"] + " provider: " + _tmp["provider_info"]["name"])
                    else:
                        formMessages.append(" - " + _tmp)
            except Exception as e:
                del form.back
                flash(str(e), category="error")
        if form.replay.data:
            del form.back
            from .dataproxy_link import control
            try:
                proxy_incidents = control.replay_incidents(form.unique_string.data, form.dataproxy.data, form.chain.data, form.witness.data)

                formMessages = ["Replay has been triggered"]
                formMessages.append("")
                formMessages.append(str(len(proxy_incidents)) + " incidents found on dataproxy")
                for _tmp in proxy_incidents:
                    if type(_tmp) == dict:
                        formMessages.append(" - " + _tmp["unique_string"] + " provider: " + _tmp["provider_info"]["name"])
                    else:
                        formMessages.append(" - " + _tmp)
            except Exception as e:
                flash(str(e), category="error")
    else:
        if filter is not None:
            form.unique_string.data = filter
        form.witness.data = "All"
        del form.back

    if form.unique_string.data is not None and not form.unique_string.data.strip() == "":
        store = factory.get_incident_storage(use=use)
        incidents = list(store.get_incidents(filter_dict=dict(
            unique_string={"$regex": ".*" + form.unique_string.data + ".*"}
        )))
        if len(formMessages) > 0:
            formMessages.append("")
        formMessages.append(str(len(incidents)) + " incidents found locally for " + str(form.unique_string.data))
        for _tmp in incidents:
            formMessages.append(" - " + _tmp["unique_string"] + " provider: " + _tmp["provider_info"]["name"])

    return render_template_menuinfo('generic.html', **locals())


@app.route("/wallet/new", methods=['GET', 'POST'])
def newwallet():
    form = NewWalletForm()
    formTitle = "Enter password to create new wallet"
    formMessage = (
        "A local wallet will be automatically created." +
        " This local wallet is encrypted with your password, and" +
        " will contain any private keys belonging to your accounts." +
        " It is important that you take the time to backup this wallet" +
        " once created!")

    if form.validate_on_submit():
        try:
            Node().wallet_create(form.password.data)
            return redirect(url_for('account_add'))
        except WalletExists as e:
            flash('There is already an open wallet.', category='error')
        except Exception as e:
            flash(e.__repr__(), category='error')

    return render_template_menuinfo('generic.html', **locals())


@app.route('/mainnet')
def mainnet():
    formMessage = "This MINT instance is connected to the peerplays mainnet (alice). Be vigilant!"
    return render_template_menuinfo('generic.html', formMessage=formMessage)


@app.route('/witnesses')
@unlocked_wallet_required
def witnesses():
    if not Config.get("advanced_features", False):
        abort(404)

    expected_version = Config.get("witnesses_versions", []).copy()

    def _forward_to_beacons():
        witnesses = Config.get("witnesses")
        responses = {}
        threads = []
        for witness in witnesses:
            def _call_beacon(_responses, _beacon, _name):
                try:
                    response = requests.get(_beacon, timeout=10)
                    if response.status_code == 200 and response.json() is not None:
                        _responses[_name] = response.json()
                    else:
                        sleep(0.2)
                        response = requests.get(_beacon, timeout=10)
                        if response.status_code == 200 and response.json() is not None:
                            _responses[_name] = response.json()
                        else:
                            _responses[_name] = "HTTP response " + str(response.status_code)
                except Exception as e:
                    try:
                        sleep(0.2)
                        response = requests.get(_beacon, timeout=10)
                        if response.status_code == 200 and response.json() is not None:
                            _responses[_name] = response.json()
                        else:
                            _responses[_name] = "HTTP response " + str(response.status_code)
                    except Exception as e:
                            _responses[_name] = "Errored, " + str(e)

            threads.append(Thread(target=_call_beacon, args=(responses, witness["url"] + "/isalive", witness["name"])))
            threads[len(threads) - 1].start()

        for i in range(len(threads)):
            threads[i].join()

        return responses

    responses = _forward_to_beacons()

    for key, value in responses.items():
        if type(value) == str:
            responses[key] = value
            continue
        ok = "NA"
        if value["queue"]["status"].get("default", None) is not None:
            ok = value["queue"]["status"]["default"]["count"]
        nok = "NA"
        if value["queue"]["status"].get("failed", None) is not None:
            nok = value["queue"]["status"]["failed"]["count"]
        scheduler = False
        if (value.get("background", None) is not None and
                value["background"].get("scheduler", None) is not None and
                value["background"]["scheduler"].get("running", None) is not None):
            scheduler = value["background"]["scheduler"]["running"]
        responses[key] = {
            "qeue": str(ok) + "/" + str(nok)
        }
        if not scheduler:
            responses[key]["scheduler"] = scheduler
        _version = (
            value["versions"]["bookiesports"] +
            "/" + value["versions"]["bos-auto"] +
            "/" + value["versions"]["bos-sync"] +
            "/" + value["versions"]["bos-incidents"] +
            "/" + value["versions"]["peerplays"]
        )
        if _version not in expected_version:
            responses[key]["deviating_versions"] = _version

    expected_version.append("<bookiesports>/<bos-auto>/<bos-sync>/<bos-incidents>/<peerplays>")
    responses["<witness_name>"] = {
        "queue": "<pending_incidents>/<failed_incidents>",
        "allowed_versions": expected_version
    }

    response_dict = {
        "time": str(datetime.now()),
        "reponses": responses
    }

    preformatted_string = json.dumps(
        response_dict,
        sort_keys=True,
        indent=4
    )

    return render_template_menuinfo('generic.html', preformatted_string=preformatted_string)


@app.route('/cancel')
@app.route('/cancel/<event_ids>')
@app.route('/cancel/<event_ids>/<chain>')
@unlocked_wallet_required
def cancel(event_ids=None, chain=None):
    if not Config.get("advanced_features", False):
        abort(404)

    if chain is None:
        chain = "beatrice"
    if event_ids is None:
        all_events = Node().getEvents("all")
        all_events = sorted(all_events, key=lambda k: k["start_time"])
        legacy_events_list = []
        legacy_events = {}
        legacy_events["details"] = []
        legacy_events["event_ids"] = ""
        for event in all_events:
            if string_to_date(event["start_time"]) < string_to_date():
                legacy_events["details"].append(
                    "event_id=" + event["id"] + "; start_time=" + event["start_time"]
                )
                legacy_events_list.append(event)

        legacy_events_list = sorted(legacy_events_list, key=lambda k: k["id"])
        for event in legacy_events_list:
            legacy_events["all_event_ids"] = legacy_events["event_ids"] + event["id"] + ","

        legacy_events["usage"] = "To cancel a specific event (or several): Dryrun with '/cancel/<comma-separated-list-of-ids>', execute with '/cancel/<comma-separated-list-of-ids>/send'"

        preformatted_string = json.dumps(
            legacy_events,
            sort_keys=True,
            indent=4
        )
    else:
        responses = []
        urls_to_call = {}
        for event_id in event_ids.split(","):
            event = Node().getEvent(event_id)
            teams = [x[1] for x in event["name"] if x[0] == 'en'][0]
            if " v " in teams:
                teams = teams.split(" v ")
            if " @ " in teams:
                teams = teams.split(" @ ")
                teams = [teams[1], teams[0]]

            id_string = event["start_time"] + "Z" \
                + '__' + [x[1] for x in event.eventgroup.sport["name"] if x[0] == 'en'][0] \
                + '__' + [x[1] for x in event.eventgroup["name"] if x[0] == 'en'][0] \
                + '__' + teams[0] \
                + '__' + teams[1]
            call = "canceled__None"

            proxies = Ping().get_status()

            for key, value in proxies.items():
                url = value["replay"] + "&manufacture=" + id_string + "__" + call
                if chain == "send":
                    url = re.sub('&only_report=True', '', url)
                urls_to_call[url] = url
            for url in urls_to_call.keys():
                response = requests.get(url, timeout=1000)
                if response.status_code == 200:
                    responses.append(requests.get(url, timeout=1000).json())

        preformatted_string = json.dumps(
            {
                "responses": responses,
                "urls_to_call": list(urls_to_call.keys())
            },
            sort_keys=True,
            indent=4
        )

    return render_template_menuinfo('generic.html', preformatted_string=preformatted_string)


@app.route('/incidents')
@app.route('/incidents/<matching>')
@app.route('/incidents/<matching>/<use>')
def show_incidents(from_date=None, to_date=None, matching=None, use="mongodb"):
    if request.args.get("matching_today", None) is not None:
        return redirect(url_for("show_incidents", matching=utils.date_to_string()[0:10]))

    if matching is not None:
        try:
            match_date = utils.string_to_date(matching[0:20])
            if from_date is None:
                from_date = match_date - timedelta(days=3)
            if to_date is None:
                to_date = match_date + timedelta(days=3)
        except InvalidRFC3339Error:
            pass

    if type(matching) == str:
        matching = matching.split(",")

    if from_date is None:
        from_date = request.args.get("from_date", None)
        if from_date is None:
            from_date = utils.date_to_string(-7)
        if type(from_date) == str:
            from_date = utils.string_to_date(from_date)
    if to_date is None:
        to_date = request.args.get("to_date", None)
        if to_date is None:
            to_date = utils.date_to_string(21)
        if type(to_date) == str:
            to_date = utils.string_to_date(to_date)

    store = None
    unresolved_events = None
    try:
        store = factory.get_incident_storage(use=use)
        unresolved_events = store.get_events(resolve=False)
    except IncidentStorageLostException:
        flash("BOS-mint could not find an incident store, or connection failed. Is a BOS-auto instance running alongside that grants access?")
        return redirect(url_for('overview'))

    events = []
    # resort for provider view
    for event in unresolved_events:
        try:
            event_scheduled = utils.string_to_date(event["id_string"][0:18])
        except InvalidRFC3339Error:
            try:
                event_scheduled = utils.string_to_date(event["id_string"][0:20])
            except InvalidRFC3339Error:
                event_scheduled = utils.string_to_date(event["id_string"][0:23])

        if event_scheduled <= to_date and event_scheduled >= from_date and\
                (matching is None or all([x.lower() in event["id_string"].lower() for x in matching])):
            store.resolve_event(event)
        else:
            continue
        for call in ["create", "in_progress", "finish", "result", "dynamic_bmgs", "canceled"]:
            try:
                incident_provider_dict = {}
                for incident in event[call]["incidents"]:
                    provider = incident["provider_info"]["name"]
                    try:
                        incident_dict = incident_provider_dict[provider]
                    except KeyError:
                        incident_provider_dict[provider] = {"incidents": [],
                                                            "replay_links": {}}
                        incident_dict = incident_provider_dict[provider]

                    incident_provider_dict[provider]["incidents"].append(incident)

                    try:
                        replay_url = Ping().get_replay_url(provider, incident, call)
                        if replay_url is not None:
                            incident_dict["replay_links"][incident["unique_string"]] = replay_url
                    except Exception as e:
                        pass

                event[call]["incidents_per_provider"] = incident_provider_dict
            except KeyError:
                pass
        event["reconstruct_string"] = get_reconstruct_string(event["id"])
        events.append(event)

    from_date = utils.date_to_string(from_date)
    to_date = utils.date_to_string(to_date)

    if use == "mongodb":
        use = "bos-auto"

    if use == "auto":
        use = "bos-auto"

    return render_template_menuinfo('showIncidents.html', **locals())


@app.route('/incidents/details/<incident_id>/<call>')
@app.route('/incidents/details/<incident_id>/<call>/<use>')
def show_incidents_per_id(incident_id=None, call=None, use="mongodb"):
    if use == "bos-auto":
        use = "mongodb"

    store = factory.get_incident_storage(use=use)
    try:
        event = store.get_event_by_id(incident_id, resolve=True)
    except EventNotFoundException:
        return jsonify("Event not found")

    for key in list(event.keys()):
        if not (key == call or key == "id" or key == "id_string"):
            event.pop(key)

    return jsonify(event)


@app.route("/event/incidents/<selectId>", methods=['get'])
@wallet_required
def event_incidents(selectId=None):
    event = Node().getEvent(selectId)
    incident_id = (event["start_time"] + "Z-" +
                   InternationalizedString.listToDict(event.eventgroup.sport["name"])["identifier"] + "-" +
                   InternationalizedString.listToDict(event.eventgroup["name"])["identifier"])

    return redirect(url_for("show_incidents", matching=incident_id))


@app.route('/overview')
@app.route('/overview/<typeName>')
@app.route('/overview/<typeName>/<identifier>')
def overview(typeName=None, identifier=None):
    # every exception will return to this view, thus this cannot break, otherwise endless loop!
    try:
        # selected ids
        selected = {}

        # same structure for all chain elements and list elements
        def buildListElements(tmpList):
            tmpList = sorted(tmpList, key=lambda k: k['toString'])
            for entry in tmpList:
                if entry['typeName'] == 'event':
                    entry['extraLink'] = [{
                        'title': 'Show incidents',
                        'link': 'event_incidents',
                        'icon': 'unhide'
                    }, {
                        'title': 'divider',
                    }, {
                        'title': 'Start/Resume',
                        'link': 'event_start',
                        'icon': 'lightning'
                    }, {
                        'title': 'Finish',
                        'link': 'event_status_update',
                        'icon': 'flag checkered'
                    }, {
                        'title': 'Freeze',
                        'link': 'event_freeze',
                        'icon': 'snowflake'
                    }, {
                        'title': 'Cancel',
                        'link': 'event_cancel',
                        'icon': 'minus circle'
                    }]
                elif entry['typeName'] == 'bettingmarketgroup':
                    entry['extraLink'] = [{
                        'title': 'Freeze',
                        'link': 'bettingmarketgroup_freeze',
                        'icon': 'snowflake'
                    }, {
                        'title': 'Resolve',
                        'link': 'bettingmarketgroup_resolve',
                        'icon': 'money'
                    }, {
                        'title': 'Cancel',
                        'link': 'bettingmarketgroup_cancel',
                        'icon': 'minus circle'
                    }]
            return tmpList

        def buildChainElement(parentId, typeName):
            tmpList = utils.getComprisedTypesGetter(typeName)(parentId)
            title = utils.getTitle(typeName)

            if typeName == 'bettingmarketgroup':
                return {
                    'list': buildListElements(tmpList),
                    'title': title,
                    'typeName': typeName,
                    'extraLink': [{
                        'title': 'Create ' + utils.getTitle('bettingmarketgrouprule'),
                        'link': 'bettingmarketgrouprule_new',
                        'icon': 'plus'
                    }, {
                        'title': 'List ' + utils.getTitle('bettingmarketgrouprule') + 's',
                        'link': 'overview',
                        'argument': ('typeName', 'bettingmarketgrouprule'),
                        'icon': 'unhide'
                    }, {
                        'title': 'Resolve ' + utils.getTitle('bettingmarketgroup') + 's',
                        'link': 'bettingmarketgroup_resolve_selectgroup',
                        'argument': ('eventId', parentId),
                        'icon': 'money'
                    }
                    ]}
            else:
                return {
                    'list': buildListElements(tmpList),
                    'title': title,
                    'typeName': typeName
                }

        # bettingmarketroule has no parent or childs
        if typeName == 'bettingmarketgrouprule' and identifier:
            flash('Betting market group rules don''t have children to display')
            return redirect(url_for('overview', typeName='bettingmarketgrouprule'))

        # build reverse chain
        reverseChain = []
        if typeName and identifier:
            # if both type name and identifier are given, its the most currently
            # selected. we also want to list all the child elements then
            tmpTypeName = utils.getChildType(typeName)
        elif typeName and not identifier:
            # if only a typename is given, the user wants to have an overview of
            # all objects of that type.
            # This method is only viable for sports, since for other elements we
            # need the parent id for a scalable select.
            # Nevertheless, the API would support it if the methods are properly
            # implementing in Node.get<typeName>s(selectedParentId) for
            # selectedParentId=None
            if typeName != 'sport' and typeName != 'bettingmarketgrouprule':
                flash('Selecting all is only available for sports and rules due to' +
                      ' performance, please specify a parent id.')
                typeName = 'sport'

            tmpTypeName = typeName
        else:
            # Nothing specified? Show sports
            tmpTypeName = 'sport'

        # reverse through all parents starting with the type given by typeName
        tmpParentIdentifier = identifier
        while tmpTypeName and not tmpTypeName == 'sport':
            # build chain element for tmpTypeName
            tmpChainElement = buildChainElement(tmpParentIdentifier,
                                                tmpTypeName)

            tmpTypeName = utils.getParentType(tmpTypeName)
            if tmpTypeName:
                selected[tmpTypeName] = tmpParentIdentifier
                tmpParentIdentifier = utils.getComprisedParentTypeGetter(
                    tmpTypeName)(tmpParentIdentifier)

            if not tmpChainElement["typeName"] == "bet":
                if isinstance(tmpChainElement, list):
                    for item in tmpChainElement:
                        reverseChain.append(item)
                else:
                    reverseChain.append(tmpChainElement)

        if tmpTypeName == 'sport':
            # sport doesnt loop through the former while,
            # thus set initial element
            sportElement = buildChainElement(None, 'sport')
            reverseChain.append(sportElement)

        reverseChain.reverse()
        listChain = reverseChain[0]
        if reverseChain:
            tmpChainElement = []
            for chainElement in reverseChain:
                if tmpChainElement:
                    tmpChainElement["nextChainElement"] = chainElement

                tmpChainElement = chainElement

        del tmpTypeName, tmpParentIdentifier

        return render_template_menuinfo('index.html', **locals())
    except Exception as e:
        app.logger.exception(e)
        if 'NumRetriesReached' in str(e):
            flash("Blockchain connection might be lost: " + str(e))
        else:
            flash(str(e))
        try:
            return render_template_menuinfo('index.html')
        except Exception as e:
            app.logger.exception(e)
            abort(500)


@app.route("/pending/discard", methods=['GET', 'POST'])
@unlocked_wallet_required
def pending_operations_discard():
    Node().discardPendingTransaction()
    flash('All pending operations have been discarded.')

    if session.get('automatic_approval', False):
        pass

    return redirect(url_for('pending_operations'))


@app.route("/pending/broadcast", methods=['POST'])
@unlocked_wallet_required
def pending_operations_broadcast():
    if ViewConfiguration.get('automatic_approval', 'enabled', False):
        if Config.get("advanced_features", default=False):
            Node().get_node().blocking = True
        else:
            flash("Automatic approval currently not supported", category="warning")

    try:
        answer = Node().broadcastPendingTransaction()

        if ViewConfiguration.get('automatic_approval', 'enabled', False):
            try:
                proposalId = answer['trx']['operation_results'][0][1]
                message = 'All pending operations have been broadcasted and the resulting proposal has been approved.'
                Node().acceptProposal(proposalId)
            except Exception as e:
                message = 'All pending operations have been broadcasted, but the resulting proposal could not be approved automatically (' + e.__class__.__name__ + ':' + str(e) + ')'
        else:
            message = 'All pending operations have been broadcasted.'
        if Node().get_node().nobroadcast:
            message += " But NoBroadcast config is set to true!"
        flash(message)
        return redirect(url_for('pending_operations'))
    except Exception as e:
        raise e
    finally:
        Node().get_node().blocking = ViewConfiguration.get('synchronous_operations', 'enabled', False)


@app.route("/pending", methods=['get'])
def pending_operations():
    form = ApprovalForm()
    form.approve.data = ViewConfiguration.get('automatic_approval', 'enabled', False)

    # construct operationsform from active transaction
    transaction = Node().getPendingTransaction()
    if transaction:
        containerList = [widgets.prepareTransactionDataForRendering(transaction)]
    del transaction

    return render_template_menuinfo("pendingOperations.html", **locals())


@app.route("/pending/automaticapproval", methods=['post'])
def automatic_approval():
    form = ApprovalForm()

    if form.validate_on_submit():
        ViewConfiguration.set('automatic_approval', 'enabled', form.approve.data)

    return redirect(url_for('pending_operations'))


@app.route("/proposals", methods=['post', 'get'])
def votable_proposals():
    proposals = Node().getAllProposals()
    for proposal in proposals:
        operations = proposal['proposed_transaction']['operations']
        if len(operations) > 0:
            operation = operations[0][1]
            if 'betting_market_group_id' in operation.keys():
                #  print('operation', operation)
                bmgId = operation['betting_market_group_id']
                #  print('bmgId', bmgId)
                bmg = Node().get_node().rpc.get_object(bmgId)
                if isinstance(bmg, dict):
                    if 'event_id' in bmg.keys():
                        event = Node().get_node().rpc.get_object(
                            bmg['event_id'])
                        proposal['event'] = str(event)

        for operation in operations:
            operation = operation[1]
            if 'resolutions' in operation.keys():                                                                                                                                                        
                resolutions = operation['resolutions']
                for resolution in resolutions:
                    bmId = resolution[0]
                    try:
                        bm = Node().get_node().rpc.get_object(bmId)
                        if isinstance(bm, dict):
                            bmName = bm['description'][0][1]
                            resolution[0] = bmName
                    except:
                        #  These print lines are for detecting a few error cases, which I couldn't reproduce, Jemshid
                        print('-------line 864, view.py, bmId')
                        print(bmId)

    if proposals:
        accountId = Node().getSelectedAccount()['id']

        containerList = widgets.prepareProposalsDataForRendering(proposals)
        containerReview = {}
        reviewedProposals = LocalProposal.getAllAsList()

        for proposal in proposals:
            if proposal['id'] in reviewedProposals:
                # if the proposal is stored in the localproposals database it already has been reviewed, but maybe
                # rejected
                containerReview[proposal['id']] = {'reviewed': True, 'approved': accountId in proposal.get('available_active_approvals')}
            elif accountId in proposal.get('available_active_approvals'):
                # already approved ones also go into the reviewed column, even without a localproposal entry
                containerReview[proposal['id']] = {'reviewed': True, 'approved': True}

    del proposals

    return render_template_menuinfo("votableProposals.html", **locals())


@app.route("/proposals/accept/<proposalId>", methods=['post', 'get'])
@unlocked_wallet_required
def votable_proposals_accept(proposalId):
    Node().acceptProposal(proposalId)
    LocalProposal.wasReviewed(proposalId)
    flash('Proposal (' + proposalId + ') has been accepted')
    return redirect(url_for('votable_proposals'))


@app.route("/proposals/reject/<proposalId>", methods=['post', 'get'])
@unlocked_wallet_required
def votable_proposals_reject(proposalId):
    try:
        Node().rejectProposal(proposalId)
    except Exception:
        pass

    LocalProposal.wasReviewed(proposalId)
    flash('Proposal (' + proposalId + ') has been rejected')
    return redirect(url_for('votable_proposals'))


def findAndProcessTranslatons(form):
    for field in form._fields.values():
        if isinstance(field, FormField) and isinstance(field.form, TranslatedFieldForm) and field.addLanguage.data:
            # an additional language line was requested
            field.translations.append_entry()
            return True

    return False


def genericNewForm(formClass, parentId=None):
    # look into https://github.com/Semantic-Org/Semantic-UI/issues/2978 for highlighting the chosen menu branch as well
    form = formClass()
    typeName = formClass.getTypeName()
    typeNameTitle = utils.getTitle(typeName)

    selectedParent = None
    default = None
    if parentId:
        selectedParent = utils.getComprisedParentByIdGetter(typeName)(parentId)
        default = {'parentId': parentId}

    form.init(selectedParent, default)

    # Which button was pressed?
    if findAndProcessTranslatons(form):
        return render_template_menuinfo("new.html", **locals())

    elif form.submit.data and form.validate_on_submit():
        # Create new sport
        operation = form.create()
        flash("A creation proposal for a new " +
              utils.getTitle(typeName) + " has been added to your shopping cart and will be" +
              " displayed with a relative id in the overview.")
        if operation is None:
            return redirect(url_for('overview'))
        else:
            return redirect(url_for('overview',
                                    typeName=operation['typeName'],
                                    identifier=operation['id']))

    return render_template_menuinfo("new.html", **locals())


@app.route("/sport/new", methods=['post', 'get'])
@unlocked_wallet_required
def sport_new():
    return genericNewForm(forms.NewSportForm)


@app.route("/eventgroup/new", methods=['post', 'get'])
@app.route("/eventgroup/new/<parentId>", methods=['post', 'get'])
@unlocked_wallet_required
def eventgroup_new(parentId=None):
    return genericNewForm(forms.NewEventGroupForm, parentId)


@app.route("/event/new/<parentId>", methods=['post', 'get'])
@unlocked_wallet_required
def event_new(parentId):
    return genericNewForm(forms.NewEventForm, parentId)


@app.route("/bettingmarketgroup/new/<parentId>", methods=['post', 'get'])
@unlocked_wallet_required
def bettingmarketgroup_new(parentId):
    return genericNewForm(forms.NewBettingMarketGroupForm, parentId)


@app.route("/bettingmarketgrouprule/new", methods=['post', 'get'])
@unlocked_wallet_required
def bettingmarketgrouprule_new(parentId=None):
    return genericNewForm(forms.NewBettingMarketGroupRuleForm)


@app.route("/bettingmarket/new/<parentId>", methods=['post', 'get'])
@unlocked_wallet_required
def bettingmarket_new(parentId):
    return genericNewForm(forms.NewBettingMarketForm, parentId)


@app.route("/bet/new", methods=['post', 'get'])
@unlocked_wallet_required
def bet_new():
    flash("Bet creation not implemented")
    return render_template_menuinfo('index.html', **locals())


def genericUpdate(formClass, selectId, removeSubmits=False, details=False):
    typeName = formClass.getTypeName()

    selectFunction = utils.getTypeGetter(typeName)
    choicesFunction = utils.getTypesGetter(typeName)

    # maybe only query the selected object, if one is preselected,
    # saves traffic currently: always query all
    parentId = None
    if selectId:
        parentId = utils.getParentTypeGetter(typeName)(selectId)
        
    form = forms.buildUpdateForm(typeName,
                                 choicesFunction(parentId),
                                 formClass,
                                 selectId,
                                 details)

    typeNameTitle = utils.getTitle(typeName)

    if not selectId:
        if form.submit.data:
            return redirect(url_for(typeName + '_update',
                                    selectId=form.select.data))
        else:
            return render_template_menuinfo("update.html", **locals())

    # update was called with given id, make sure it exists
    if selectId:
        selectedObject = selectFunction(selectId)

    # help file present?
    if os.path.isfile("bos_mint/static/img/help/" + typeName + ".png"):
        help_file = "../../static/img/help/" + typeName + ".png"
    elif os.path.isfile(os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            'static',
            'img',
            'help',
            typeName + ".png"
    )):
        help_file = "../../static/img/help/" + typeName + ".png"

    form.init(selectedObject)
 
    # user wants to add language?
    if findAndProcessTranslatons(form):
        return render_template_menuinfo("update.html", **locals())

    # first request? populate selected object
    if not form.submit.data:
        # preselect
        form.select.data = selectId
        form.fill(selectedObject)

    if removeSubmits:
        # remove submit button
        # todo: disable any and all user input
        typeNameAction = 'Details of'

        for fieldName in form._fields.copy().keys():
            # avoid mutating original set while iterating
            field = form._fields[fieldName]

            if isinstance(field, SubmitField):
                delattr(form, fieldName)
            elif isinstance(field, FormField):
                for subfieldName in field.form._fields.keys():
                    subField = field.form._fields[subfieldName]
                    if isinstance(subField, SubmitField):
                        delattr(field.form, subfieldName)

        return render_template_menuinfo("update.html", **locals())

    else:
        typeNameAction = 'Update'

    if form.validate_on_submit():
        # user submitted, wants to change
        # all data was entered correctly, validate and update sport
        proposal = form.update(selectedObject['id'])
        flash("An update proposal  for " + utils.getTitle(typeName) +
              " (id=" + selectId + ") has been added to your shopping cart.")
        return redirect(utils.processNextArgument(
            request.args.get('next'), 'index'))

    return render_template_menuinfo("update.html", **locals())

def proposal_awaiting_approval(pendingUpdate=None):
    proposals = Node().getAllProposals()
    # Should not update proposal when the proposal of same id is pending for approval
    approve = False
    for proposal in proposals:
        for operation in proposal['proposed_transaction']['operations']:
            for d in operation[1:]:
                for k,v in d.items():
                    if k in pendingUpdate.keys() and v in pendingUpdate.values():
                       approve = True
        del operation

    return approve


@app.route("/sport/update", methods=['post', 'get'])
@app.route("/sport/update/<selectId>", methods=['post', 'get'])
@unlocked_wallet_required
def sport_update(selectId=None):
    dict1 = {'sport_id' : selectId}
    if not proposal_awaiting_approval(dict1):
        return genericUpdate(forms.NewSportForm, selectId)
    else:
        flash("Current Sport update is not allowed, Proposal is pending for approval")
        return genericUpdate(forms.SportFormDetails, selectId, True, True)


@app.route("/eventgroup/update", methods=['post', 'get'])
@app.route("/eventgroup/update/<selectId>", methods=['post', 'get'])
@unlocked_wallet_required
def eventgroup_update(selectId=None):
    dict1 = {'event_group_id' : selectId}
    if not proposal_awaiting_approval(dict1):
        formClass = forms.NewEventGroupForm
        return genericUpdate(formClass, selectId)
    else:
        flash("Current EventGroup update is not allowed, Proposal is pending for approval")
        formClass = forms.EventGroupFormDetails
        return genericUpdate(formClass, selectId, True, True) 


@app.route("/event/update", methods=['post', 'get'])
@app.route("/event/update/<selectId>", methods=['post', 'get'])
@unlocked_wallet_required
def event_update(selectId=None):
    dict1 = {'event_id' : selectId}
    if not proposal_awaiting_approval(dict1):
        formClass = forms.NewEventForm
        return genericUpdate(formClass, selectId)
    else:
        flash("Current Event update is not allowed, Proposal is pending for approval")
        formClass = forms.EventFormDetails
        return genericUpdate(formClass, selectId, True, True)


@app.route("/event/finish", methods=['post', 'get'])
@app.route("/event/finish/<selectId>", methods=['post', 'get'])
@unlocked_wallet_required
def event_status_update(selectId=None):
    formClass = forms.EventStatusForm
    return genericUpdate(formClass, selectId)


@app.route("/bettingmarketgroup/update", methods=['post', 'get'])
@app.route("/bettingmarketgroup/update/<selectId>", methods=['post', 'get'])
@unlocked_wallet_required
def bettingmarketgroup_update(selectId=None):
    dict1 = {'betting_market_group_id' : selectId}
    if not proposal_awaiting_approval(dict1):
        formClass = forms.NewBettingMarketGroupForm
        return genericUpdate(formClass, selectId)
    else:
        flash("Current BettingMarketGroup update is not allowed, Proposal is pending for approval")
        formClass = forms.BettingMarketGroupFormDetails
        return genericUpdate(formClass, selectId, True, True)

@app.route("/bettingmarket/update", methods=['post', 'get'])
@app.route("/bettingmarket/update/<selectId>", methods=['post', 'get'])
@unlocked_wallet_required
def bettingmarket_update(selectId=None):
    dict1 = {'betting_market_id' : selectId}
    if not proposal_awaiting_approval(dict1):
        formClass = forms.NewBettingMarketForm
        return genericUpdate(formClass, selectId)
    else:
        flash("Current BettingMarket update is not allowed, Proposal is pending for approval")
        formClass = forms.BettingMarketFormDetails
        return genericUpdate(formClass, selectId, True, True)

@app.route("/bettingmarketgrouprule/update", methods=['post', 'get'])
@app.route("/bettingmarketgrouprule/update/<selectId>", methods=['post', 'get'])
@unlocked_wallet_required
def bettingmarketgrouprule_update(selectId=None):
    dict1 = {'betting_market_rules_id' : selectId}
    if not proposal_awaiting_approval(dict1):
        formClass = forms.NewBettingMarketGroupRuleForm
        return genericUpdate(formClass, selectId)
    else:
        flash("Current BettingMarketGroupRule update is not allowed, Proposal is pending for approval")
        formClass = forms.BettingMarketGroupRuleFormDetails
        return genericUpdate(formClass, selectId, True, True)

@app.route("/sport/details/<selectId>")
def sport_details(selectId):
    return genericUpdate(forms.SportFormDetails, selectId, True, True)


@app.route("/eventgroup/details/<selectId>")
def eventgroup_details(selectId):
    formClass = forms.EventGroupFormDetails
    return genericUpdate(formClass, selectId, True, True)


@app.route("/event/details/<selectId>")
def event_details(selectId):
    formClass = forms.EventFormDetails
    return genericUpdate(formClass, selectId, True, True)


@app.route("/bettingmarketgroup/details/<selectId>")
def bettingmarketgroup_details(selectId):
    formClass = forms.BettingMarketGroupFormDetails
    return genericUpdate(formClass, selectId, True, True)


@app.route("/bettingmarketgrouprule/details/<selectId>")
def bettingmarketgrouprule_details(selectId):
    formClass = forms.BettingMarketGroupRuleFormDetails
    return genericUpdate(formClass, selectId, True, True)


@app.route("/bettingmarket/details/<selectId>")
def bettingmarket_details(selectId):
    formClass = forms.BettingMarketFormDetails
    return genericUpdate(formClass, selectId, True, True)


@app.route("/event/start/<selectId>", methods=['post', 'get'])
@unlocked_wallet_required
def event_start(selectId):
    if Node().getEvent(selectId).get("status") == "in_progress":
        flash("Event is already in 'in_progress' state")
    else:
        Node().startEvent(selectId)
    return redirect(utils.processNextArgument(
                    request.args.get('next'), 'index'))


@app.route("/event/freeze/<selectId>", methods=['post', 'get'])
@unlocked_wallet_required
def event_freeze(selectId=None):
    Node().freezeEvent(selectId)
    return redirect(utils.processNextArgument(
                    request.args.get('next'), 'index'))


@app.route("/event/cancel/<selectId>", methods=['post', 'get'])
@unlocked_wallet_required
def event_cancel(selectId=None):
    Node().cancelEvent(selectId)
    return redirect(utils.processNextArgument(
                    request.args.get('next'), 'index'))


@app.route("/bettingmarketgroup/freeze/<selectId>", methods=['post', 'get'])
@unlocked_wallet_required
def bettingmarketgroup_freeze(selectId=None):
    Node().freezeBettingMarketGroup(selectId)
    return redirect(utils.processNextArgument(
                    request.args.get('next'), 'index'))


@app.route("/bettingmarketgroup/resolve/selectgroup/<eventId>", methods=['get', 'post'])
def bettingmarketgroup_resolve_selectgroup(eventId=None):
    form = BettingMarketGroupResolveForm("Betting market group resolution - select a group to resolve")

    selectedEvent = Node().getEvent(eventId)

    form.initEvents(selectedEvent['event_group_id'], eventId)
    form.event.render_kw = {"disabled": True}

    form.initGroups(selectedEvent)
#     form.bettingmarketgroup.render_kw = {'onclick': "document.getElementById('bettingMarketGroupResolveForm').submit()"}

    if not form.submit.data:
        form.fillEvent(selectedEvent)
        form.event.render_kw = {"disabled": True}

    if form.validate_on_submit():
        return redirect(url_for('bettingmarketgroup_resolve',
                        selectId=form.bettingmarketgroup.data))

    return render_template_menuinfo("generic.html", **locals())


@app.route("/bettingmarketgroup/resolve/<selectId>", methods=['post', 'get'])
@unlocked_wallet_required
def bettingmarketgroup_resolve(selectId=None):
    form = BettingMarketGroupResolveForm()
    bettingMarketGroupId = selectId
    selectedBMG = Node().getBettingMarketGroup(bettingMarketGroupId)

    form.initEvents(selectedBMG.event['event_group_id'])
    form.initGroups(selectedBMG['event_id'])

    if not form.submit.data:
        form.fillMarkets(bettingMarketGroupId)

        form.fillEvent(Node().getEvent(selectedBMG['event_id']))
        form.event.render_kw = {"disabled": True}

        form.bettingmarketgroup.data = bettingMarketGroupId
        form.bettingmarketgroup.render_kw = {"disabled": True}

        # todo: this is weird, i dont understand why data is filled
        for market in form.bettingmarkets:
            market.resolution.data = None

    if form.validate_on_submit():
        resultList = []
        for market in form.bettingmarkets:
            resultList.append([market.identifier.data, market.resolution.data])
        Node().resolveBettingMarketGroup(bettingMarketGroupId, resultList)
        flash("An update proposal to resolve Betting market group " + str(bettingMarketGroupId) +
              " has been added to your shopping cart.")
        return redirect(utils.processNextArgument(
                        request.args.get('next'), 'index'))

    return render_template_menuinfo("generic.html", **locals())


@app.route("/bettingmarketgroup/unfreeze/<selectId>", methods=['post', 'get'])
@unlocked_wallet_required
def bettingmarketgroup_unfreeze(selectId=None):
    Node().unfreezeBettingMarketGroup(selectId)
    return redirect(utils.processNextArgument(
                    request.args.get('next'), 'index'))


@app.route("/bettingmarketgroup/unfreeze/<selectId>", methods=['post', 'get'])
@unlocked_wallet_required
def bettingmarketgroup_cancel(selectId=None):
    Node().cancelBettingMarketGroup(selectId)
    return redirect(utils.processNextArgument(
                    request.args.get('next'), 'index'))



# -*- coding: utf-8 -*-


class LanguageNotFoundException(BaseException):
    pass


class InternationalizedString(object):
    LANGUAGES = [
        ("display_name", "Display Name"),
        ("sen", "Short English"),
        ("ab", "Abkhazian"),
        ("aa", "Afar"),
        ("af", "Afrikaans"),
        ("ak", "Akan"),
        ("sq", "Albanian"),
        ("am", "Amharic"),
        ("ar", "Arabic"),
        ("an", "Aragonese"),
        ("hy", "Armenian"),
        ("as", "Assamese"),
        ("av", "Avaric"),
        ("ae", "Avestan"),
        ("ay", "Aymara"),
        ("az", "Azerbaijani"),
        ("bm", "Bambara"),
        ("ba", "Bashkir"),
        ("eu", "Basque"),
        ("be", "Belarusian"),
        ("bn", "Bengali (Bangla)"),
        ("bh", "Bihari"),
        ("bi", "Bislama"),
        ("bs", "Bosnian"),
        ("br", "Breton"),
        ("bg", "Bulgarian"),
        ("my", "Burmese"),
        ("ca", "Catalan"),
        ("ch", "Chamorro"),
        ("ce", "Chechen"),
        ("ny", "Chichewa, Chewa, Nyanja"),
        ("zh", "Chinese"),
        ("zh_Hans", "Chinese (Simplified)"),
        ("zh_Hant", "Chinese (Traditional)"),
        ("cv", "Chuvash"),
        ("kw", "Cornish"),
        ("co", "Corsican"),
        ("cr", "Cree"),
        ("hr", "Croatian"),
        ("cs", "Czech"),
        ("da", "Danish"),
        ("dv", "Divehi, Dhivehi, Maldivian"),
        ("nl", "Dutch"),
        ("dz", "Dzongkha"),
        ("en", "English"),
        ("eo", "Esperanto"),
        ("et", "Estonian"),
        ("ee", "Ewe"),
        ("fo", "Faroese"),
        ("fj", "Fijian"),
        ("fi", "Finnish"),
        ("fr", "French"),
        ("ff", "Fula, Fulah, Pulaar, Pular"),
        ("gl", "Galician"),
        ("gd", "Gaelic (Scottish)"),
        ("gv", "Gaelic (Manx)"),
        ("ka", "Georgian"),
        ("de", "German"),
        ("gr", "Greek"),
        ("kl", "Greenlandic"),
        ("gn", "Guarani"),
        ("gu", "Gujarati"),
        ("ht", "Haitian Creole"),
        ("ha", "Hausa"),
        ("he", "Hebrew"),
        ("hz", "Herero"),
        ("hi", "Hindi"),
        ("ho", "Hiri Motu"),
        ("hu", "Hungarian"),
        ("is", "Icelandic"),
        ("io", "Ido"),
        ("ig", "Igbo"),
        ("id", "Indonesian"),
        ("in", "Indonesian"),
        ("ia", "Interlingua"),
        ("ie", "Interlingue"),
        ("iu", "Inuktitut"),
        ("ik", "Inupiak"),
        ("ga", "Irish"),
        ("it", "Italian"),
        ("ja", "Japanese"),
        ("jv", "Javanese"),
        ("kl", "Kalaallisut, Greenlandic"),
        ("kn", "Kannada"),
        ("kr", "Kanuri"),
        ("ks", "Kashmiri"),
        ("kk", "Kazakh"),
        ("km", "Khmer"),
        ("ki", "Kikuyu"),
        ("rw", "Kinyarwanda (Rwanda)"),
        ("rn", "Kirundi"),
        ("ky", "Kyrgyz"),
        ("kv", "Komi"),
        ("kg", "Kongo"),
        ("ko", "Korean"),
        ("ku", "Kurdish"),
        ("kj", "Kwanyama"),
        ("lo", "Lao"),
        ("la", "Latin"),
        ("lv", "Latvian (Lettish)"),
        ("li", "Limburgish ( Limburger)"),
        ("ln", "Lingala"),
        ("lt", "Lithuanian"),
        ("lu", "Luga-Katanga"),
        ("lg", "Luganda, Ganda"),
        ("lb", "Luxembourgish"),
        ("gv", "Manx"),
        ("mk", "Macedonian"),
        ("mg", "Malagasy"),
        ("ms", "Malay"),
        ("ml", "Malayalam"),
        ("mt", "Maltese"),
        ("mi", "Maori"),
        ("mr", "Marathi"),
        ("mh", "Marshallese"),
        ("mo", "Moldavian"),
        ("mn", "Mongolian"),
        ("na", "Nauru"),
        ("nv", "Navajo"),
        ("ng", "Ndonga"),
        ("nd", "Northern Ndebele"),
        ("ne", "Nepali"),
        ("no", "Norwegian"),
        ("nb", "Norwegian bokm��l"),
        ("nn", "Norwegian nynorsk"),
        ("ii", "Nuosu"),
        ("oc", "Occitan"),
        ("oj", "Ojibwe"),
        ("cu", "Old Church Slavonic, Old Bulgarian"),
        ("or", "Oriya"),
        ("om", "Oromo (Afaan Oromo)"),
        ("os", "Ossetian"),
        ("pi", "P��li"),
        ("ps", "Pashto, Pushto"),
        ("fa", "Persian (Farsi)"),
        ("pl", "Polish"),
        ("pt", "Portuguese"),
        ("pa", "Punjabi (Eastern)"),
        ("qu", "Quechua"),
        ("rm", "Romansh"),
        ("ro", "Romanian"),
        ("ru", "Russian"),
        ("se", "Sami"),
        ("sm", "Samoan"),
        ("sg", "Sango"),
        ("sa", "Sanskrit"),
        ("sr", "Serbian"),
        ("sh", "Serbo-Croatian"),
        ("st", "Sesotho"),
        ("tn", "Setswana"),
        ("sn", "Shona"),
        ("ii", "Sichuan Yi"),
        ("sd", "Sindhi"),
        ("si", "Sinhalese"),
        ("ss", "Siswati"),
        ("sk", "Slovak"),
        ("sl", "Slovenian"),
        ("so", "Somali"),
        ("nr", "Southern Ndebele"),
        ("es", "Spanish"),
        ("su", "Sundanese"),
        ("sw", "Swahili (Kiswahili)"),
        ("ss", "Swati"),
        ("sv", "Swedish"),
        ("tl", "Tagalog"),
        ("ty", "Tahitian"),
        ("tg", "Tajik"),
        ("ta", "Tamil"),
        ("tt", "Tatar"),
        ("te", "Telugu"),
        ("th", "Thai"),
        ("bo", "Tibetan"),
        ("ti", "Tigrinya"),
        ("to", "Tonga"),
        ("ts", "Tsonga"),
        ("tr", "Turkish"),
        ("tk", "Turkmen"),
        ("tw", "Twi"),
        ("ug", "Uyghur"),
        ("uk", "Ukrainian"),
        ("ur", "Urdu"),
        ("uz", "Uzbek"),
        ("ve", "Venda"),
        ("vi", "Vietnamese"),
        ("vo", "Volap��k"),
        ("wa", "Wallon"),
        ("cy", "Welsh"),
        ("wo", "Wolof"),
        ("fy", "Western Frisian"),
        ("xh", "Xhosa"),
        ("yi", "Yiddish"),
        ("ji", "Yiddish"),
        ("yo", "Yoruba"),
        ("za", "Zhuang, Chuang"),
        ("zu", "Zulu"),
        ("unknown", "Unknown language"),
    ]
    UNKNOWN = "Unknown"

    def __init__(self, country, text):
        '''
        Constructor
        '''
        self.country = country
        self.text = text

        if self.country not in [
            x[0] for x in InternationalizedString.LANGUAGES
        ] and not country == self.UNKNOWN:
            raise LanguageNotFoundException

    def getForm(self):
        from bos_mint.forms import InternationalizedStringForm
        lng = InternationalizedStringForm()
        lng.country = self.country
        lng.text = self.text
        return lng

    @classmethod
    def listToDict(cls, listOfIStrings):
        return {x[0]: x[1] for x in listOfIStrings}

    @classmethod
    def getChoices(cls):
        return [
            (x[0], x[0] + " - " + x[1])
            for x in InternationalizedString.LANGUAGES
        ]

    @classmethod
    def parseToList(cls, fieldListOfInternationalizedString):
        from bos_mint.forms import TranslatedFieldForm
        from wtforms import FormField

        istrings = list()

        if (
            isinstance(
                fieldListOfInternationalizedString, FormField) and
            isinstance(
                fieldListOfInternationalizedString.form, TranslatedFieldForm
            )
        ):
            fieldListOfInternationalizedString = \
                fieldListOfInternationalizedString.form.translations

        for ffield in fieldListOfInternationalizedString.entries:
            istrings.append([ffield.data["country"], ffield.data["text"]])

        return istrings



import click


@click.group()
def main():
    pass


@main.command()
@click.option("--port", type=int, default=5000)
@click.option("--host", type=str, default="localhost")
def start(port, host):
    from bos_mint import Config
    from bos_mint.web import app
    app.run(debug=Config.get("debug"), port=port, host=host)


if __name__ == "__main__":
    main()



import os
from flask_assets import Bundle, Environment
from . import app

# Assets
webassets = Environment(app)
webassets.load_path = [
    os.path.join(os.path.dirname(__file__), "static"),
    os.path.join(os.path.dirname(__file__), "bower_components")
]
webassets.manifest = 'cache' if not app.debug else False
webassets.cache = not app.debug
webassets.debug = True  # app.debug

js_main = Bundle(
    "js/src/main.js",
    output="js/main.js"
)

css_main = Bundle(
    "css/src/styles.css",
    output="css/main.css"
)

js_libs = Bundle(
    "jquery/dist/jquery.min.js",
    "semantic-ui/dist/semantic.min.js",
    output="js/libs.js"
)

css_libs = Bundle(
    "font-awesome/web-fonts-with-css/css/fontawesome.min.css",
    "semantic-ui/dist/semantic.min.css",
    output="css/libs.css"
)

webassets.register('js_main', js_main)
webassets.register('js_libs', js_libs)
webassets.register('css_main', css_main)
webassets.register('css_libs', css_libs)



from . import db


class LocalProposal(db.Model):
    proposelIdentifier = db.Column(db.String(128), primary_key=True)
    reviewed = db.Column(db.Boolean, default=False)

    @classmethod
    def getAllAsList(cls):
        lpList = LocalProposal.query.all()
        return [x.proposelIdentifier for x in lpList]

    @classmethod
    def wasReviewed(cls, proposalId):
        # check if it exists
        lp = LocalProposal.query.get(proposalId)

        if lp is None:
            lp = LocalProposal(proposelIdentifier=proposalId, reviewed=True)
            db.session.add(lp)

        db.session.commit()


class ViewConfiguration(db.Model):
    name = db.Column(db.String(128), primary_key=True)
    key = db.Column(db.String(128), primary_key=True)
    value = db.Column(db.String(128))

    @classmethod
    def set(cls, name, key, value):
        if type(value) == bool:
            if value:
                value = 'True'
            else:
                value = 'False'

        # check if it exists
        vc = ViewConfiguration.query.filter_by(name=name, key=key).first()

        if vc is None:
            vc = ViewConfiguration(name=name, key=key, value=value)
            db.session.add(vc)
        else:
            vc.value = value

        db.session.commit()

    @classmethod
    def get(cls, name, key, default):
        # check if it exists
        vc = ViewConfiguration.query.filter_by(name=name, key=key).first()

        if vc is None:
            return default
        else:
            if type(default) is bool:
                if vc.value == 'True':
                    return True
                else:
                    return False

            return vc.value



from flask import Flask, redirect, url_for, session, current_app, request
from flaskext.markdown import Markdown
from . import app, assets, views, config, db, forms
from .utils import strfdelta
from .node import Node, ApiServerDown

# # Blueprints
# from .referral import referral
# from .wallet import wallet
# app.register_blueprint(referral, url_prefix="/referral")
# app.register_blueprint(wallet, url_prefix="/wallet")

# Config markdown
markdown = Markdown(
    app,
    extensions=['meta',
                'tables'
                ],
    safe_mode=True,
    output_format='html4',
)


@app.template_filter('datetime')
def _jinja2_filter_datetime(date, fmt=None):
    if not date:
        return "unkown"
    if fmt:
        return strfdelta(date, fmt)
    else:
        return strfdelta(date, '{days} days {hours} hours')



from datetime import datetime, timedelta
import strict_rfc3339
import pytz


def date_to_string(date_object=None):
    """ rfc3339 conform string represenation of a date
        can also be given as str YYYY-mm-dd HH:MM:SS """
    if type(date_object) == int:
        if date_object < 365:
            date_object = datetime.utcnow() + timedelta(days=date_object)
        else:
            date_object = datetime.utcfromtimestamp(date_object)
    if type(date_object) == float:
        date_object = datetime.utcfromtimestamp(date_object)
    if type(date_object) == str:
        try:
            date_object = datetime.strptime(date_object + "+0000",
                                            '%Y-%m-%d %H:%M:%S%z')
        except ValueError:
            date_object = string_to_date(date_object)
    if not date_object:
        return strict_rfc3339.now_to_rfc3339_utcoffset()
    else:
        if date_object.tzinfo is None:
            return strict_rfc3339.timestamp_to_rfc3339_utcoffset(
                pytz.UTC.localize(date_object).timestamp())
        else:
            return strict_rfc3339.timestamp_to_rfc3339_utcoffset(
                date_object.timestamp())


def string_to_date(date_string=None):
    """ assumes rfc3339 conform string and creates date object """
    if date_string is None:
        date_time_object = datetime.utcnow()
        return pytz.utc.localize(date_time_object)
    if type(date_string) == str:
        if len(date_string) == 8:
            date_string = date_string[0:4] + "-" + date_string[4:6] + "-" + date_string[6:8] + "T00:00:00Z"
        if len(date_string) == 10:
            date_string = date_string + "T00:00:00Z"
        if len(date_string) == 19:
            date_string = date_string + "Z"
        if len(date_string) == 18 and date_string[10:11] == "t":
            date_string = date_string[0:10] + "T" + date_string[11:13] + ":" + date_string[13:15] + ":" + date_string[15:17] + "Z"
        date_time_object = datetime.utcfromtimestamp(
            strict_rfc3339.rfc3339_to_timestamp(date_string))
        return pytz.utc.localize(date_time_object)
    raise Exception("Only string covnersion supported")



from flask import url_for
from peerplaysbase import operationids
import os.path
from .node import Node
from . import tostring


class RenderTemplateWidget(object):
    """
        Base template for every widget
        Enables the possibility of rendering a template
         inside a template with run time options
    """
    template = 'appbuilder/general/widgets/render.html'
    template_args = None

    def __init__(self, **kwargs):
        self.template_args = kwargs

    def __repr__(self, **kwargs):
        return self.__call__(**kwargs)

    def __call__(self, **kwargs):
        from flask.globals import _request_ctx_stack
        ctx = _request_ctx_stack.top
        jinja_env = ctx.app.jinja_env

        template = jinja_env.get_template(self.template)
        args = self.template_args.copy()
        args.update(kwargs)
        return template.render(args)


class OperationsContainerWidget(RenderTemplateWidget):
    template = 'widgets/operationContainer.html'

    def __init__(self, **kwargs):
        if not kwargs.get('operations'):
            kwargs['operations'] = []

        super(OperationsContainerWidget, self).__init__(**kwargs)

    @property
    def id(self):
        return self.template_args['id']

    def addOperation(self, operationId, data):
        ow = OperationWidget(
            operationId=operationId,
            operationName=operationids.getOperationNameForId(operationId),
            data=data
        )
        self.template_args['operations'].append(ow)


class OperationWidget(RenderTemplateWidget):
    template = None

    def __init__(self, **kwargs):
        super(OperationWidget, self).__init__(**kwargs)

        name = operationids.getOperationNameForId(kwargs['operationId'])
        file = 'operation_' + name + '.html'

        if kwargs['operationId'] == 22:
            self.template = os.path.join('widgets', 'operation_proposal.html')

            # add child operations
            operation = kwargs['data']
            self.template_args['title'] = 'Proposal'
            self.template_args['listItems'] = [
                ('Fee paying account', operation['fee_paying_account']),
                ('Expiration time', operation['expiration_time'])
            ]

            for tmpOp in operation['proposed_ops']:
                self.addOperation(tmpOp['op'][0], tmpOp['op'][1])
        elif os.path.isfile(os.path.join(
                'bos_mint',
                'templates',
                'widgets',
                file
        )):
            self.template = os.path.join('widgets', file)
        elif os.path.isfile(os.path.join(
                'templates',
                'widgets',
                file
        )):
            self.template = os.path.join('widgets', file)
        elif os.path.isfile(os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                'templates',
                'widgets',
                file
        )):
            self.template = os.path.join('widgets', file)
        else:
            self.template = 'widgets/operation_unknown.html'

    def addOperation(self, operationId, data):
        if not self.template_args.get('operations'):
            self.template_args['operations'] = []

        ow = OperationWidget(
            operationId=operationId,
            operationName=operationids.getOperationNameForId(operationId),
            data=data
        )
        self.template_args['operations'].append(ow)


def prepareProposalsDataForRendering(proposals, accountId=None):
    tmpList = []
    for proposal in proposals:
        # ensure the parent expiration time is the shortest time
        if proposal['expiration_time'] < proposal['proposed_transaction']['expiration']:
            raise Exception('Expiration times are differing')

        tmpListItems = []
        if proposal.get('expiration_time'):
            tmpListItems.append(('Expiration time', proposal['expiration_time']))
        if proposal.get('review_period_time'):
            tmpListItems.append(('Review period time', proposal['review_period_time']))
        if proposal.get('available_active_approvals'):
            tmpListItems.append(('Available active approvals', [
                tostring.toString(x) for x in Node().getAccounts(proposal['available_active_approvals'])]))
        if proposal.get('required_active_approvals'):
            approvalAccounts = Node().getAccounts(proposal['required_active_approvals'])
            # special handling for witness account
            accountList = []
            for account in approvalAccounts:
                if account["id"] == "1.2.1":
                    for authAccount in account["active"]["account_auths"]:
                        accountList.append(
                            tostring.toString(
                                Node().getAccount(authAccount[0])
                            )
                        )
                else:
                    accountList.append(tostring.toString(x))

            tmpListItems.append(('Required approvals', accountList))
        if proposal.get('event'):
            tmpListItems.append(('Event', str(proposal.get('event'))))

        ocw = OperationsContainerWidget(
            title='Proposal ' + proposal['id'],
            id=proposal['id'],
            listItems=tmpListItems,
            buttonNegative='Reject',
            buttonPositive='Approve',
            buttonNegativeURL=url_for('votable_proposals_reject',
                                      proposalId=proposal['id']),
            buttonPositiveURL=url_for('votable_proposals_accept',
                                      proposalId=proposal['id'])
        )

        for operation in proposal['proposed_transaction']['operations']:
            ocw.addOperation(operation[0], operation[1])

        tmpList.append(ocw)

    return tmpList


def prepareTransactionDataForRendering(transaction):
    # ensure the parent expiration time is the shortest time
    if transaction.proposal_expiration < transaction.proposal_review:
        raise Exception('Expiration times are differing')

    ocw = OperationsContainerWidget(
        title='Current transaction details',
        listItems=[
            ('Proposer', transaction.proposer),
            ('Expiration time', transaction.proposal_expiration)
        ],
        buttonNegative='Discard',
        buttonPositive='Broadcast',
        buttonNegativeURL=url_for('pending_operations_discard'),
        buttonPositiveURL=url_for('pending_operations_broadcast'),
    )

    for operation in transaction.get_parent()['operations']:
        ocw.addOperation(operation[0], operation[1])

    return ocw
